package build

import (
	"fmt"
	"strconv"

	"git.akyoto.dev/cli/q/src/build/arch/x64"
	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/config"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/go/color/ansi"
)

// Function represents a function.
type Function struct {
	Name      string
	File      *File
	Head      token.List
	Body      token.List
	Variables map[string]*Variable
	Assembler asm.Assembler
	Error     error
}

// Compile turns a function into machine code.
func (f *Function) Compile() {
	if config.Verbose {
		ansi.Bold.Println(f.Name)
	}

	start := 0
	groupLevel := 0

	for i, t := range f.Body {
		if start == i && (t.Kind == token.NewLine || t.Kind == token.BlockStart || t.Kind == token.BlockEnd) {
			start = i + 1
			continue
		}

		switch t.Kind {
		case token.NewLine:
			if groupLevel > 0 {
				continue
			}

			if start != -1 {
				instruction := f.Body[start:i]
				err := f.CompileInstruction(instruction)

				if err != nil {
					f.Error = err
					return
				}

				start = -1
			}

			start = i + 1

		case token.GroupStart:
			groupLevel++

		case token.GroupEnd:
			groupLevel--

		case token.BlockStart:
			// Add scope

		case token.BlockEnd:
			// Remove scope
		}
	}

	f.Assembler.Return()

	if config.Verbose {
		fmt.Println()
		ansi.Bold.Println(f.Name + ".asm")

		for _, x := range f.Assembler.Instructions {
			ansi.Dim.Print("│ ")
			fmt.Print(x.Mnemonic.String())

			if x.Data != nil {
				fmt.Print(" " + x.Data.String())
			}

			fmt.Print("\n")
		}
	}
}

// CompileInstruction compiles a single instruction.
func (f *Function) CompileInstruction(line token.List) error {
	if config.Verbose {
		ansi.Dim.Print("│ ")
		fmt.Println(line)
	}

	if line[0].Kind == token.Keyword {
		switch line[0].Text() {
		case "return":
			f.Assembler.Return()
		}

		return nil
	}

	if line[0].Kind == token.Identifier {
		if len(line) < 2 {
			return fmt.Errorf("error to be implemented")
		}

		if line[1].Kind == token.Define {
			name := line[0].Text()
			value := line[2:]

			if len(value) == 0 {
				return fmt.Errorf("error to be implemented")
				// return errors.New(errors.MissingAssignmentValue, f.File.Path, f.File.Tokens, f.Cursor)
			}

			if config.Verbose {
				ansi.Dim.Printf("├── var ")
				fmt.Println(name, "=", value)
			}

			f.Variables[name] = &Variable{
				Name:    name,
				Value:   value,
				IsConst: true,
			}

			return nil
		}

		switch line[0].Text() {
		case "syscall":
			paramTokens := line[2 : len(line)-1]
			start := 0
			i := 0
			var parameters []token.List

			for i < len(paramTokens) {
				if paramTokens[i].Kind == token.Separator {
					parameters = append(parameters, paramTokens[start:i])
					start = i + 1
				}

				i++
			}

			if i != start {
				parameters = append(parameters, paramTokens[start:i])
			}

			for i, list := range parameters {
				switch list[0].Kind {
				case token.Identifier:
					name := list[0].Text()
					variable, exists := f.Variables[name]

					if !exists {
						panic("Unknown identifier " + name)
					}

					if !variable.IsConst {
						panic("Not implemented yet")
					}

					n, _ := strconv.Atoi(variable.Value[0].Text())
					f.Assembler.MoveRegisterNumber(x64.SyscallArgs[i], uint64(n))

				case token.Number:
					value := list[0].Text()
					n, _ := strconv.Atoi(value)
					f.Assembler.MoveRegisterNumber(x64.SyscallArgs[i], uint64(n))

				default:
					panic("Unknown expression")
				}
			}

			f.Assembler.Syscall()
		}
	}

	return nil
}

// String returns the function name.
func (f *Function) String() string {
	return f.Name
}
